@@ -304,6 +304,12 @@ void cairo_context::paint()
     check_object_status_and_throw_exception(*this);
 }
 
+void cairo_context::paint(double opacity)
+{
+    cairo_paint_with_alpha(cairo_.get(), opacity);
+    check_object_status_and_throw_exception(*this);
+}
+
 void cairo_context::set_pattern(cairo_pattern const& pattern)
 {
     cairo_set_source(cairo_.get(), pattern.pattern());
@@ -489,6 +495,18 @@ void cairo_context::add_text(glyph_positions const& pos,
 
 }
 
+void cairo_context::push_group()
+{
+    cairo_push_group(cairo_.get());
+    check_object_status_and_throw_exception(*this);
+}
+
+void cairo_context::pop_group()
+{
+    cairo_pop_group_to_source(cairo_.get());
+    check_object_status_and_throw_exception(*this);
+}
+
 cairo_face_manager::cairo_face_manager(std::shared_ptr<font_library> font_library)
   : font_library_(font_library)
 {
